#include "BF_ML.h"
#include <stdio.h>
//#include <iostream>

namespace BF
{
	CML::CML(int _numFreq, int _samplingRate, double _soundSpeed, double _regFactor) :
		BF(_numFreq, _samplingRate, _soundSpeed), 
		regFactor(_regFactor)
	{
	}

	CML::~CML()
	{
	}

	void
	CML::updateSources(const std::vector<SourceLocation> &sourcePositions)
	{
/*== source from MLBeam.m
	%----------------------------------
	%  Covariance
	%----------------------------------
	wnd = hamming(NFFT)';

	NFrame = floor((BlockLength-(FrameLength-FrameShift))/FrameShift);
	NBlock = floor(NData/BlockLength);

	CC = zeros(NFreq,M,M);

	BlockOrigin = BlockLength*(NoiseBlock-1);
	ZB = zeros(NFreq,M,NFrame);
	for frame = 1:NFrame
		t1 = BlockOrigin + FrameShift*(frame-1)+1;
		t2 = BlockOrigin + FrameShift*(frame-1)+FrameLength;
		for m=1:M
			Z = fft( z(m,t1:t2).*wnd, NFFT );
			ZB(:,m,frame) = reshape(Z(1:NFreq),NFreq,1,1);
		end
	end

	for freq=1:NFreq
		ZZ = reshape(ZB(freq,:,:),M,NFrame);
		KK(freq,:,:) = reshape(ZZ*ZZ',1,M,M);
	end

	%----------------------------
	%  Target loop
	%----------------------------

	for tl = 1:NTargetLoc
		target = TargetLocation(tl);
		%----------------------------
		%  Beamformer Design
		%----------------------------
		for k=1:NFFT/2+1
			a = reshape(AA(k,:,target),M,1);
			a = a/norm(a);
			K = reshape(KK(k,:,:),M,M);
			K = K + norm(K,'fro')*RegFactor*eye(M);
			invK = inv(K);
			WW(:,k) = invK*a/(a'*invK*a);
		end

	end

	for k=1:NFFT/2+1
		W = reshape(WW(:,k),1,M);
		H = exp(j*(k-1)*pi);				% Shift by NFFT/2;
		for m=1:M
			FF(k,m) = H*conj(W(m));                 % conj for y = w'x in freq. domain
		end
	end
*/

		const int numSrc = sourcePositions.size();

		/*== 伝達関数を再計算する ==*/
		std::vector<Matrix> newHs;
		this->BF::getTransformFunctions(sourcePositions, newHs);

		/*== 分離行列を計算する ==*/
		std::vector<Matrix> newGs;
		for (int i = 0; i < this->numFreq; i++) {
			/* 伝達関数 */
			const Matrix &A = newHs[i];
			int numMic = A.getRowCount();

			/* 相関行列 */
			Matrix K = covarianceMatrices[i];
			///K = K.conjugated();
			Matrix KK = K + K.frobeniusNorm() * this->regFactor * Matrix::I(numMic);
			KK = KK.pseudoInverse();

			/* ML Beamformer */
			Matrix G(numSrc, numMic);
			for (int j = 0; j < numSrc; j++) {
				Matrix AA = A.getColumn(j);
				///AA = AA.conjugated();
				Matrix U, S, V;
				AA.singularValueDecomposition(U, S, V);
				Complex Hnorm2 = S.getMaximumValue();
				if (abs(Hnorm2) < 1.0e-10) {
//std::cout << "norm(A)[" << i << ", " << j << "] " << Hnorm2 << " : less than 1.0e-6" << std::endl;
					Hnorm2 = 1.0e-10;
				}
				AA = AA * (1.0 / Hnorm2);

				Matrix KA = KK * AA;
				Matrix AtKA = AA.transposed() * KA;
				Complex AtKAsclr = AtKA.scalar();
				if (abs(AtKAsclr) < 1.0e-10) {
//std::cout << "(A.ctranspose * inv(K) * A)[" << i << ", " << j << "] " << AtKAsclr << " : less than 1.0e-15" << std::endl;
					AtKAsclr = 1.0e-10;
				}
				Matrix WW = KA * (1.0 / AtKAsclr); // WW [rows:numMic, cols:1]"

				Complex H = exp(Complex(0.0, 1.0) * (double)i * M_PI);
				for (int k = 0; k < numMic; k++) {
					Complex W = WW.getElementAt(k, 0);
					G.setElementAt(j, k, H * std::conj(W));
				}
			}
			newGs.push_back(G);
		}

		/*== 結果の格納 ==*/
		this->sourcePositions = sourcePositions;
		newHs.swap(this->transformFunctions);
		newGs.swap(this->separationMatrices);
	}

	void
	CML::adaptSeparationMatrices()
	{
		/* do nothing */
	}
}
